@@ -2,12 +2,17 @@ import datetime
 from io import BytesIO
 
 from botocore.auth import (
-    _host_from_url, _get_body_as_dict, BaseSigner,
-    SIGNED_HEADERS_BLACKLIST, UNSIGNED_PAYLOAD
+    SIGNED_HEADERS_BLACKLIST,
+    STREAMING_UNSIGNED_PAYLOAD_TRAILER,
+    UNSIGNED_PAYLOAD,
+    BaseSigner,
+    _get_body_as_dict,
+    _host_from_url,
 )
-from botocore.compat import awscrt, HTTPHeaders, parse_qs, urlsplit, urlunsplit
-from botocore.utils import percent_encode_sequence
+from botocore.compat import HTTPHeaders, awscrt, parse_qs, urlsplit, urlunsplit
 from botocore.exceptions import NoCredentialsError
+from botocore.utils import percent_encode_sequence
+
 
 class CrtSigV4Auth(BaseSigner):
     REQUIRES_REGION = True
@@ -27,6 +32,11 @@ class CrtSigV4Auth(BaseSigner):
         self._region_name = region_name
         self._expiration_in_seconds = None
 
+    def _is_streaming_checksum_payload(self, request):
+        checksum_context = request.context.get('checksum', {})
+        algorithm = checksum_context.get('request_algorithm')
+        return isinstance(algorithm, dict) and algorithm.get('in') == 'trailer'
+
     def add_auth(self, request):
         if self.credentials is None:
             raise NoCredentialsError()
@@ -46,7 +56,9 @@ class CrtSigV4Auth(BaseSigner):
             secret_access_key=self.credentials.secret_key,
             session_token=self.credentials.token)
 
-        if self._should_sha256_sign_payload(request):
+        if self._is_streaming_checksum_payload(request):
+            explicit_payload = STREAMING_UNSIGNED_PAYLOAD_TRAILER
+        elif self._should_sha256_sign_payload(request):
             if existing_sha256:
                 explicit_payload = existing_sha256
             else:
@@ -73,7 +85,7 @@ class CrtSigV4Auth(BaseSigner):
             signed_body_value=explicit_payload,
             signed_body_header_type=body_header,
             expiration_in_seconds=self._expiration_in_seconds,
-            )
+        )
         crt_request = self._crt_request_from_aws_request(request)
         future = awscrt.auth.aws_sign_request(crt_request, signing_config)
         future.result()
@@ -171,12 +183,17 @@ class CrtS3SigV4Auth(CrtSigV4Auth):
         if sign_payload is not None:
             return sign_payload
 
-        # We require that both content-md5 be present and https be enabled
+        # We require that both a checksum be present and https be enabled
         # to implicitly disable body signing. The combination of TLS and
-        # content-md5 is sufficiently secure and durable for us to be
+        # a checksum is sufficiently secure and durable for us to be
         # confident in the request without body signing.
+        checksum_header = 'Content-MD5'
+        checksum_context = request.context.get('checksum', {})
+        algorithm = checksum_context.get('request_algorithm')
+        if isinstance(algorithm, dict) and algorithm.get('in') == 'header':
+            checksum_header = algorithm['name']
         if not request.url.startswith('https') or \
-                'Content-MD5' not in request.headers:
+                checksum_header not in request.headers:
             return True
 
         # If the input is streaming we disable body signing by default.
@@ -229,7 +246,9 @@ class CrtSigV4AsymAuth(BaseSigner):
             secret_access_key=self.credentials.secret_key,
             session_token=self.credentials.token)
 
-        if self._should_sha256_sign_payload(request):
+        if self._is_streaming_checksum_payload(request):
+            explicit_payload = STREAMING_UNSIGNED_PAYLOAD_TRAILER
+        elif self._should_sha256_sign_payload(request):
             if existing_sha256:
                 explicit_payload = existing_sha256
             else:
@@ -256,7 +275,7 @@ class CrtSigV4AsymAuth(BaseSigner):
             signed_body_value=explicit_payload,
             signed_body_header_type=body_header,
             expiration_in_seconds=self._expiration_in_seconds,
-            )
+        )
         crt_request = self._crt_request_from_aws_request(request)
         future = awscrt.auth.aws_sign_request(crt_request, signing_config)
         future.result()
@@ -312,6 +331,11 @@ class CrtSigV4AsymAuth(BaseSigner):
     def _get_existing_sha256(self, request):
         return request.headers.get('X-Amz-Content-SHA256')
 
+    def _is_streaming_checksum_payload(self, request):
+        checksum_context = request.context.get('checksum', {})
+        algorithm = checksum_context.get('request_algorithm')
+        return isinstance(algorithm, dict) and algorithm.get('in') == 'trailer'
+
     def _should_sha256_sign_payload(self, request):
         # Payloads will always be signed over insecure connections.
         if not request.url.startswith('https'):
@@ -374,6 +398,7 @@ class CrtS3SigV4AsymAuth(CrtSigV4AsymAuth):
         # Always add X-Amz-Content-SHA256 header
         return True
 
+
 class CrtSigV4AsymQueryAuth(CrtSigV4AsymAuth):
     DEFAULT_EXPIRES = 3600
     _SIGNATURE_TYPE = awscrt.auth.AwsSignatureType.HTTP_REQUEST_QUERY_PARAMS
@@ -491,6 +516,9 @@ class CrtSigV4QueryAuth(CrtSigV4Auth):
         query_dict = dict(
             [(k, v[0]) for k, v in
              parse_qs(url_parts.query, keep_blank_values=True).items()])
+        if request.params:
+            query_dict.update(request.params)
+            request.params = {}
         # The spec is particular about this.  It *has* to be:
         # https://<endpoint>?<operation params>&<auth params>
         # You can't mix the two types of params together, i.e just keep doing
